/*
 *   Copyright 2012 Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

/**
 * @file   bitmap_malloc.c
 * @author Ilias Pliotas, Ioannis Koutras
 * @date   September, 2012
 * @brief  malloc() implementation for bitmap-organized raw blocks
 */

#include "bitmap/bitmap.h"
#include "bitmap/bitmap_rb.h"
#include "bitmap/bitmap_other.h"

#include "trace.h"

#ifdef HAVE_LOCKS
#include <pthread.h>
#endif /* HAVE_LOCKS */

#ifdef WITH_ALLOCATOR_STATS
#include "dmmlib/dmmlib.h"
#endif /* WITH_ALLOCATOR_STATS */

/**
 * Returns a memory block from a bitmap-organized raw block
 *
 * @param  raw_block The pointer of the raw block.
 * @param  req_size  The requested memory size.
 * @retval           The address to serve the request.
 * @retval NULL      No available memory space.
 */
void * bitmap_malloc(raw_block_header_t *raw_block, size_t req_size) { 
    bitmap_rb_t *rb_header;
    size_t cells, stop, i, found;
    void *ret;
    chunk_header_t *chunk_address;
    BMAP_EL_TYPE mask1, mask2, *bmap_p;

    rb_header = (bitmap_rb_t *)((char *)raw_block + sizeof(raw_block_header_t));

    BMAP_EL_TYPE temp1[rb_header->elements], temp2[rb_header->elements];

    ret = NULL;

    req_size += CHUNK_HDR_SIZE;

    if(req_size % rb_header->bytes_per_cell > 0) {
        cells = req_size / rb_header->bytes_per_cell + 1;
    } else {
        cells = req_size / rb_header->bytes_per_cell;
    }

    // FIXME The current implementation does not allow checking cells which
    // lie on more than two consecutive bitmap vector elements
    if(cells > 2 * BMAP_EL_SIZE_BITS) {
        return NULL;
    }

    stop = prev_pow2(cells);

#ifdef HAVE_LOCKS
    pthread_mutex_lock(&raw_block->mutex);
#endif /* HAVE_LOCKS */

    copy_array(temp1, &rb_header->bmap, rb_header->elements);

    // perform bitwise shift & and operations in the BMAP_EL_TYPE arrays
    for(i = 1; i < stop; i <<= 1) {
        copy_array(temp2, temp1, rb_header->elements);
        shift_array(temp2, i, rb_header->elements);
        add_arrays(temp1, temp2, rb_header->elements);
    }
    if(stop < cells) {
        copy_array(temp2, temp1, rb_header->elements);
        shift_array(temp2, cells - stop, rb_header->elements);
        add_arrays(temp1, temp2, rb_header->elements);
    }

    found = 0;

    bmap_p = &rb_header->bmap;
    for(i = 0; i < rb_header->elements; ++i) {
        found = __builtin_ffsl(temp1[i]);
        if(found) {
            // Check if one or two elements have to be modified
            if(found + (cells - 1) <= BMAP_EL_SIZE_BITS) {
                mask1 = ~make_bit_mask(found, cells);
            } else {
                mask1 = ~make_bit_mask(found, BMAP_EL_SIZE_BITS - found + 1);
                mask2 = ~make_bit_mask((size_t) 1,
                        cells - (BMAP_EL_SIZE_BITS - found) - 1);
                bmap_p[i + 1] &= mask2;
            }
            bmap_p[i] &= mask1;

            // Calculate the pointer to the chunk to be retrieved
            chunk_address = (chunk_header_t *)((char *)rb_header +
                    sizeof(bitmap_rb_t) +
                    (rb_header->elements - 1) * BMAP_EL_SIZE +
                    (i * BMAP_EL_SIZE_BITS + found - 1) *
                    rb_header->bytes_per_cell);
            chunk_address->num_of_cells = cells;
#ifdef REQUEST_SIZE_INFO
            chunk_address->requested_size = req_size - CHUNK_HDR_SIZE;
#endif /* REQUEST_SIZE_INFO */

            ret = (void *)((char *)chunk_address + CHUNK_HDR_SIZE);

            TRACE_1("dmmlib - malloc - allocated %zu bytes"
                    " at bitmap raw block %p\n",
                    cells * rb_header->bytes_per_cell,
                    (void *)raw_block);

#ifdef WITH_ALLOCATOR_STATS
            systemallocator.dmm_stats.total_mem_allocated +=
                cells * rb_header->bytes_per_cell;
            systemallocator.dmm_stats.live_objects++;
            systemallocator.dmm_stats.num_malloc++;
            TRACE_1("dmmlib - global allocated memory: %zu bytes\n",
                    systemallocator.dmm_stats.total_mem_allocated);
#ifdef REQUEST_SIZE_INFO
            systemallocator.dmm_stats.total_mem_requested += req_size -
                CHUNK_HDR_SIZE;
            TRACE_1("dmmlib - global requested memory: %zu bytes\n",
                    systemallocator.dmm_stats.total_mem_requested);
#endif /* REQUEST_SIZE_INFO */
#endif /* WITH_ALLOCATOR_STATS */

#ifdef WITH_RAWBLOCK_STATS
            raw_block->dmm_stats.total_mem_allocated +=
                cells * rb_header->bytes_per_cell;
            raw_block->dmm_stats.live_objects++;
            raw_block->dmm_stats.num_malloc++;
            TRACE_1("dmmlib - total allocated memory: %zu bytes\n",
                    raw_block->dmm_stats.total_mem_allocated);
#ifdef REQUEST_SIZE_INFO
            raw_block->dmm_stats.total_mem_requested += req_size -
                CHUNK_HDR_SIZE;
            TRACE_1("dmmlib - total requested memory: %zu bytes\n",
                    raw_block->dmm_stats.total_mem_requested);
#endif /* REQUEST_SIZE_INFO */
#endif /* WITH_RAWBLOCK_STATS */

            break;
        }
    }

#ifdef HAVE_LOCKS
    pthread_mutex_unlock(&raw_block->mutex);
#endif /* HAVE_LOCKS */

    return ret;
} 
