/*
 *   Copyright 2011 Institute of Communication and Computer Systems (ICCS) 
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 */

#include "freelist/split.h"

#include "freelist/block_header_funcs.h"
#include "freelist/linked_lists/add_block.h"

void split(freelist_rb_t *raw_block, block_header_t *ptr, size_t req_size) {
    size_t new_size;
    size_t min_split_size;
    block_header_t *new_block;
#ifdef COALESCE_AFTER_SPLIT
    size_t max_coal_size;
    block_header_t *next_block;
    size_t current_next_size;

    next_block = get_dlnext(allocator, ptr);
#endif /* COALESCE_AFTER_SPLIT */

    /* Check what would be the size of the new block if we split the current
     * one.
     * Note: new_size is a size_t, so compare first in order to prevent an
     * underflow.
     */
    if(get_size(ptr) > req_size + HEADER_SIZE) {
        new_size = get_size(ptr) - req_size - HEADER_SIZE;
    } else {
        new_size = 0;
    }

#ifdef SPLITTING_FIXED
    min_split_size = MIN_SPLITTING_SIZE;
#endif /* SPLITTING_FIXED */
#ifdef SPLITTING_VARIABLE
    min_split_size = heap->dmm_knobs.min_split_size;
#endif /* SPLITTING_VARIABLE */

    if(new_size < min_split_size) {
        return;
    }

    new_block = (block_header_t *)((char *)ptr + req_size + HEADER_SIZE);

    if(raw_block->border_ptr == ptr) {
        raw_block->border_ptr = new_block;
    }

    /* Resize the previous, to be used block */
    set_size_and_used(raw_block, ptr, req_size);

#ifdef COALESCE_AFTER_SPLIT
    /* Try to coalesce with the next block if it is free */
#ifdef COALESCING_FIXED
    max_coal_size = MAX_COALESCE_SIZE;
#endif /* COALESCING_FIXED */
#ifdef COALESCING_VARIABLE
    max_coal_size = heap->dmm_knobs.max_coalesce_size;
#endif /* COALESCING_VARIABLE */

    if(next_block != (memory_block_ptr) NULL) {
        if(is_free(next_block) == true) {
            current_next_size = new_size + get_size(next_block) +
                HEADER_SIZE;
            if(current_next_size <= max_coal_size) {
                /* it's ok to coalesce the new split block with the next
                 * free block
                 */
                remove_block_from_lists(&next_block, heap);
                if(allocator->border_ptr == next_block) {
                    allocator->border_ptr = ptr;
                }
                new_size = current_next_size;
            }
        }
    }

#endif /* COALESCE_AFTER_SPLIT */

    set_size_and_free(raw_block, new_block, new_size);

#ifdef FIFO_SORT_POLICY
    add_block(&new_block, &heap->free_list_head, &heap->free_list_tail);
#else /* FIFO_SORT_POLICY */
    add_block(raw_block, new_block);
#endif /* FIFO_SORT_POLICY */

}
